#include <dht11.h>
#include <LiquidCrystal_I2C.h>
#include <Ultrasonic.h>
#include <DHT.h>
#include <Keypad.h>

#define DHT11PIN 7
#define Password_Length 5
#define BUZZER_PIN PD3 
#define LED_POWER_PIN PD6

const int buzzer = 3;

char Data[Password_Length]; 
char Master[Password_Length] = "1234"; 
byte data_count = 0, master_count = 0;
bool Pass_is_good;
char customKey;

const byte ROWS = 4; 
const byte COLS = 3; 

char hexaKeys[ROWS][COLS] = {
  {'1', '2', '3'},
  {'4', '5', '6'},
  {'7', '8', '9'},
  {'*', '0', '#'}
};

byte rowPins[ROWS] = {13, 12, 11, 10}; 
byte colPins[COLS] = {5, 4, 2}; 

Keypad customKeypad = Keypad(makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS); 

LiquidCrystal_I2C lcd(0x27, 16, 2);
DHT dht(DHT11PIN, DHT11);

const int trigPin = 9;
const int echoPin = 8;

Ultrasonic ultrasonic(trigPin, echoPin);

int measurePin = A0;

unsigned int samplingTime = 280;
unsigned int deltaTime = 40;
unsigned int sleepTime = 9680;

float voMeasured = 0;
float calcVoltage = 0;
float dustDensity = 0;

float distance = 0;
float motionThreshold = 20; 

// Variables to store temperature and humidity readings
float temperature = 0;
float humidity = 0;

// Threshold values for fire detection
const float temperatureThreshold = 45.0; 
const float dustThreshold = 0.45;

void activateBuzzer() {
  tone(buzzer, 1000);
  delay(1000);
}

void deactivateBuzzer() {
  noTone(buzzer);
  delay(1000);
}

void getTemperatureData() {
  temperature = dht.readTemperature();

  Serial.print("Temperature: ");
  Serial.print(temperature);
  Serial.print(" °C\t");
  Serial.println();
}

void setupADC() {
  ADMUX = 0;
  ADMUX |= (1 << MUX0);
  ADMUX |= (1 << REFS0);

  ADCSRA = 0;
  ADCSRA |= (7 << ADPS0);
  ADCSRA |= (1 << ADEN);
}


uint16_t read_adc()
{
  ADCSRA |= (1 << ADSC);
  while (!(ADCSRA & (1 << ADIF)));
  return ADC;
} 



void getDustData() {
  PORTD &= ~(1 << LED_POWER_PIN);
  delayMicroseconds(samplingTime);

  voMeasured = read_adc();

  delayMicroseconds(deltaTime);
  PORTD |= (1 << LED_POWER_PIN);
  delayMicroseconds(sleepTime);

  calcVoltage = voMeasured *(5.0 / 1024);
  dustDensity = 0.17 * calcVoltage - 0.1;

  if ( dustDensity < 0)
  {
    dustDensity = 0.00;
  }

  Serial.print("Dust Density: ");
  Serial.println(dustDensity);
}

void getMotionData() {
  distance = ultrasonic.read();

  Serial.print("Distance: ");
  Serial.print(distance);
  Serial.println(" cm");
}

void doFireDetection() {
  // Check if temperature or humidity exceeds the thresholds
  if (temperature > temperatureThreshold || dustDensity > dustThreshold) {
    // Fire detected
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Fire detected!");
    activateBuzzer();
    delay(1000);
    readPassword();
  } else {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("No fire");
    lcd.setCursor(0, 1);
    lcd.print("detected!");
  }
}

void doMotionDetection() {
  // Check if the measured distance is less than the motion threshold
  if (distance < motionThreshold) {
    // Motion detected
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("Motion detected!");
    activateBuzzer();
    delay(1000);
    readPassword();
  } else {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("No motion");
    lcd.setCursor(0, 1);
    lcd.print("detected!");
  } 
}

void clearData(){
  while(data_count !=0){
    Data[data_count--] = 0; 
  }
  return;
}

void readPassword() {
  lcd.clear();
  lcd.setCursor(0,0);
  lcd.print("Enter Password:");

  while (data_count != Password_Length - 1) {
    customKey = customKeypad.getKey();
    if (customKey) {
      Data[data_count] = customKey; 
      lcd.setCursor(data_count,1); 
      lcd.print(Data[data_count]); 
      data_count++; 
    }
  }

  delay(1000);

  if(data_count == Password_Length - 1){
    lcd.clear();

    if(!strcmp(Data, Master)){
      lcd.print("Correct");
      deactivateBuzzer();
      delay(1000);
      clearData();  
    } else {
      lcd.print("Incorrect");
      delay(1000);
      clearData();  
      readPassword();
    }
  }
}

void setup()
{
  lcd.init();
  lcd.backlight();
  lcd.begin(16,2);
  Serial.begin(9600);

  // Initialize the DHT sensor
  dht.begin();
  setupADC();

  DDRD |= (1 << BUZZER_PIN);
  DDRD |= (1 << LED_POWER_PIN);
}

void loop()
{

  getTemperatureData();
  getDustData();

  doFireDetection();
  delay(2000);

  getMotionData();
  doMotionDetection();

  delay(1000);
}